# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.block_storage.v3 import group as _group
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import envvars
from openstackclient.i18n import _


def _format_group(group: _group.Group) -> tuple[Sequence[str], Iterable[Any]]:
    columns = (
        'id',
        'status',
        'name',
        'description',
        'group_type',
        'volume_types',
        'availability_zone',
        'created_at',
        'volumes',
        'group_snapshot_id',
        'source_group_id',
    )
    column_headers = (
        'ID',
        'Status',
        'Name',
        'Description',
        'Group Type',
        'Volume Types',
        'Availability Zone',
        'Created At',
        'Volumes',
        'Group Snapshot ID',
        'Source Group ID',
    )

    # TODO(stephenfin): Consider using a formatter for volume_types since it's
    # a list
    return (
        column_headers,
        utils.get_item_properties(
            group,
            columns,
        ),
    )


class CreateVolumeGroup(command.ShowOne):
    """Create a volume group.

    Generic volume groups enable you to create a group of volumes and manage
    them together.

    Generic volume groups are more flexible than consistency groups. Currently
    volume consistency groups only support consistent group snapshot. It
    cannot be extended easily to serve other purposes. A project may want to
    put volumes used in the same application together in a group so that it is
    easier to manage them together, and this group of volumes may or may not
    support consistent group snapshot. Generic volume group solve this problem.
    By decoupling the tight relationship between the group construct and the
    consistency concept, generic volume groups can be extended to support other
    features in the future.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        # This is a bit complicated. We accept two patterns: a legacy pattern
        #
        #   volume group create \
        #     <volume-group-type> <volume-type> [<volume-type>...]
        #
        # and the modern approach
        #
        #   volume group create \
        #     --volume-group-type <volume-group-type>
        #     --volume-type <volume-type>
        #    [--volume-type <volume-type> ...]
        #
        # Because argparse doesn't properly support nested exclusive groups, we
        # use two groups: one to ensure users don't pass <volume-group-type> as
        # both a positional and an option argument and another to ensure users
        # don't pass <volume-type> this way. It's a bit weird but it catches
        # everything we care about.
        source_parser = parser.add_mutually_exclusive_group()
        # we use a different name purely so we can issue a deprecation warning
        source_parser.add_argument(
            'volume_group_type_legacy',
            metavar='<volume_group_type>',
            nargs='?',
            help=argparse.SUPPRESS,
        )
        volume_types_parser = parser.add_mutually_exclusive_group()
        # We need to use a separate dest
        # https://github.com/python/cpython/issues/101990
        volume_types_parser.add_argument(
            'volume_types_legacy',
            metavar='<volume_type>',
            nargs='*',
            default=[],
            help=argparse.SUPPRESS,
        )
        source_parser.add_argument(
            '--volume-group-type',
            metavar='<volume_group_type>',
            help=_('Volume group type to use (name or ID)'),
        )
        volume_types_parser.add_argument(
            '--volume-type',
            metavar='<volume_type>',
            dest='volume_types',
            action='append',
            default=[],
            help=_(
                'Volume type(s) to use (name or ID) '
                '(required with --volume-group-type)'
            ),
        )
        source_parser.add_argument(
            '--source-group',
            metavar='<source-group>',
            help=_(
                'Existing volume group to use (name or ID) '
                '(supported by --os-volume-api-version 3.14 or later)'
            ),
        )
        source_parser.add_argument(
            '--group-snapshot',
            metavar='<group-snapshot>',
            help=_(
                'Existing group snapshot to use (name or ID) '
                '(supported by --os-volume-api-version 3.14 or later)'
            ),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Name of the volume group.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of a volume group.'),
        )
        parser.add_argument(
            '--availability-zone',
            metavar='<availability-zone>',
            help=_(
                'Availability zone for volume group. '
                '(not available if creating group from source)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if parsed_args.volume_group_type_legacy:
            msg = _(
                "Passing volume group type and volume types as positional "
                "arguments is deprecated. Use the --volume-group-type and "
                "--volume-type option arguments instead."
            )
            self.log.warning(msg)

        volume_group_type = (
            parsed_args.volume_group_type
            or parsed_args.volume_group_type_legacy
        )
        volume_types = parsed_args.volume_types[:]
        volume_types.extend(parsed_args.volume_types_legacy)

        if volume_group_type:
            if not sdk_utils.supports_microversion(volume_client, '3.13'):
                msg = _(
                    "--os-volume-api-version 3.13 or greater is required to "
                    "support the 'volume group create' command"
                )
                raise exceptions.CommandError(msg)
            if not volume_types:
                msg = _(
                    "--volume-types is a required argument when creating a "
                    "group from group type."
                )
                raise exceptions.CommandError(msg)

            volume_group_type_id = volume_client.find_group_type(
                volume_group_type, ignore_missing=False
            ).id
            volume_types_ids = [
                volume_client.find_type(vt, ignore_missing=False).id
                for vt in volume_types
            ]

            group = volume_client.create_group(
                group_type=volume_group_type_id,
                volume_types=volume_types_ids,
                name=parsed_args.name,
                description=parsed_args.description,
                availability_zone=parsed_args.availability_zone,
            )

            group = volume_client.get_group(group.id)
            return _format_group(group)

        else:
            if not sdk_utils.supports_microversion(volume_client, '3.14'):
                msg = _(
                    "--os-volume-api-version 3.14 or greater is required to "
                    "support the 'volume group create "
                    "[--source-group|--group-snapshot]' command"
                )
                raise exceptions.CommandError(msg)
            if (
                parsed_args.source_group is None
                and parsed_args.group_snapshot is None
            ):
                msg = _(
                    "Either --source-group <source_group> or "
                    "'--group-snapshot <group_snapshot>' needs to be "
                    "provided to run the 'volume group create "
                    "[--source-group|--group-snapshot]' command"
                )
                raise exceptions.CommandError(msg)
            if parsed_args.availability_zone:
                msg = _(
                    "'--availability-zone' option will not work "
                    "if creating group from source."
                )
                self.log.warning(msg)

            source_group_id = None
            if parsed_args.source_group:
                source_group_id = volume_client.find_group(
                    parsed_args.source_group, ignore_missing=False
                ).id
            group_snapshot_id = None
            if parsed_args.group_snapshot:
                group_snapshot_id = volume_client.find_group_snapshot(
                    parsed_args.group_snapshot, ignore_missing=False
                ).id

            group = volume_client.create_group_from_source(
                group_snapshot_id=group_snapshot_id,
                source_group_id=source_group_id,
                name=parsed_args.name,
                description=parsed_args.description,
            )
            group = volume_client.get_group(group.id)
            return _format_group(group)


class DeleteVolumeGroup(command.Command):
    """Delete a volume group.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Name or ID of volume group to delete'),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help=_(
                'Delete the volume group even if it contains volumes. '
                'This will delete any remaining volumes in the group.'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.13'):
            msg = _(
                "--os-volume-api-version 3.13 or greater is required to "
                "support the 'volume group delete' command"
            )
            raise exceptions.CommandError(msg)

        group = volume_client.find_group(
            parsed_args.group,
            ignore_missing=False,
        )

        volume_client.delete_group(group, delete_volumes=parsed_args.force)


class SetVolumeGroup(command.ShowOne):
    """Update a volume group.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Name or ID of volume group.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name for group.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New description for group.'),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--enable-replication',
            action='store_true',
            dest='enable_replication',
            default=None,
            help=_(
                'Enable replication for group. '
                '(supported by --os-volume-api-version 3.38 or above)'
            ),
        )
        type_group.add_argument(
            '--disable-replication',
            action='store_false',
            dest='enable_replication',
            help=_(
                'Disable replication for group. '
                '(supported by --os-volume-api-version 3.38 or above)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.13'):
            msg = _(
                "--os-volume-api-version 3.13 or greater is required to "
                "support the 'volume group set' command"
            )
            raise exceptions.CommandError(msg)

        group = volume_client.find_group(
            parsed_args.group,
            ignore_missing=False,
        )

        if parsed_args.enable_replication is not None:
            if not sdk_utils.supports_microversion(volume_client, '3.38'):
                msg = _(
                    "--os-volume-api-version 3.38 or greater is required to "
                    "support the '--enable-replication' or "
                    "'--disable-replication' options"
                )
                raise exceptions.CommandError(msg)

            if parsed_args.enable_replication:
                volume_client.enable_group_replication(group)
            else:
                volume_client.disable_group_replication(group)

        kwargs = {}

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name

        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description

        if kwargs:
            group = volume_client.update_group(group, **kwargs)

        return _format_group(group)


class ListVolumeGroup(command.Lister):
    """Lists all volume groups.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            dest='all_projects',
            action='store_true',
            default=envvars.boolenv('ALL_PROJECTS'),
            help=_('Shows details for all projects (admin only).'),
        )
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.13'):
            msg = _(
                "--os-volume-api-version 3.13 or greater is required to "
                "support the 'volume group list' command"
            )
            raise exceptions.CommandError(msg)

        groups = list(
            volume_client.groups(all_projects=parsed_args.all_projects)
        )

        column_headers = (
            'ID',
            'Status',
            'Name',
        )
        columns = (
            'id',
            'status',
            'name',
        )

        return (
            column_headers,
            (utils.get_item_properties(a, columns) for a in groups),
        )


class ShowVolumeGroup(command.ShowOne):
    """Show detailed information for a volume group.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Name or ID of volume group.'),
        )
        parser.add_argument(
            '--volumes',
            action='store_true',
            dest='show_volumes',
            default=None,
            help=_(
                'Show volumes included in the group. '
                '(supported by --os-volume-api-version 3.25 or above)'
            ),
        )
        parser.add_argument(
            '--no-volumes',
            action='store_false',
            dest='show_volumes',
            help=_(
                'Do not show volumes included in the group. '
                '(supported by --os-volume-api-version 3.25 or above)'
            ),
        )
        parser.add_argument(
            '--replication-targets',
            action='store_true',
            dest='show_replication_targets',
            default=None,
            help=_(
                'Show replication targets for the group. '
                '(supported by --os-volume-api-version 3.38 or above)'
            ),
        )
        parser.add_argument(
            '--no-replication-targets',
            action='store_false',
            dest='show_replication_targets',
            help=_(
                'Do not show replication targets for the group. '
                '(supported by --os-volume-api-version 3.38 or above)'
            ),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.13'):
            msg = _(
                "--os-volume-api-version 3.13 or greater is required to "
                "support the 'volume group show' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.show_volumes is not None:
            if not sdk_utils.supports_microversion(volume_client, '3.25'):
                msg = _(
                    "--os-volume-api-version 3.25 or greater is required to "
                    "support the '--(no-)volumes' option"
                )
                raise exceptions.CommandError(msg)

        if parsed_args.show_replication_targets is not None:
            if not sdk_utils.supports_microversion(volume_client, '3.38'):
                msg = _(
                    "--os-volume-api-version 3.38 or greater is required to "
                    "support the '--(no-)replication-targets' option"
                )
                raise exceptions.CommandError(msg)

        group = volume_client.find_group(
            parsed_args.group,
            ignore_missing=False,
        )

        if parsed_args.show_volumes is not None:
            group = volume_client.get_group(
                group.id, list_volume=parsed_args.show_volumes
            )

        # TODO(stephenfin): Show replication targets
        return _format_group(group)


class FailoverVolumeGroup(command.Command):
    """Failover replication for a volume group.

    This command requires ``--os-volume-api-version`` 3.38 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Name or ID of volume group to failover replication for.'),
        )
        parser.add_argument(
            '--allow-attached-volume',
            action='store_true',
            dest='allow_attached_volume',
            default=False,
            help=_('Allow group with attached volumes to be failed over.'),
        )
        parser.add_argument(
            '--disallow-attached-volume',
            action='store_false',
            dest='allow_attached_volume',
            default=False,
            help=_('Disallow group with attached volumes to be failed over.'),
        )
        parser.add_argument(
            '--secondary-backend-id',
            metavar='<backend_id>',
            help=_('Secondary backend ID.'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.38'):
            msg = _(
                "--os-volume-api-version 3.38 or greater is required to "
                "support the 'volume group failover' command"
            )
            raise exceptions.CommandError(msg)

        group = volume_client.find_group(
            parsed_args.group,
            ignore_missing=False,
        )

        volume_client.failover_group_replication(
            group,
            allowed_attached_volume=parsed_args.allow_attached_volume,
            secondary_backend_id=parsed_args.secondary_backend_id,
        )



#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy

from openstackclient.object.v1 import object as obj
from openstackclient.tests.unit.object.v1 import fakes as object_fakes


class TestObjectList(object_fakes.TestObjectV1):
    columns = ('Name',)
    datalist = ((object_fakes.object_name_2,),)

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = obj.ListObject(self.app, None)

    def test_object_list_objects_no_options(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            object_fakes.container_name,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (object_fakes.object_name_1,),
            (object_fakes.object_name_2,),
        )
        self.assertEqual(datalist, tuple(data))

    def test_object_list_objects_prefix(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--prefix',
            'floppy',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('prefix', 'floppy'),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name_2,
            prefix='floppy',
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_delimiter(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--delimiter',
            '=',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('delimiter', '='),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name_2,
            delimiter='=',
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_marker(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--marker',
            object_fakes.object_name_2,
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('marker', object_fakes.object_name_2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name_2,
            marker=object_fakes.object_name_2,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_end_marker(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--end-marker',
            object_fakes.object_name_2,
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('end_marker', object_fakes.object_name_2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name_2,
            end_marker=object_fakes.object_name_2,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_limit(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--limit',
            '2',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('limit', 2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name_2,
            limit=2,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_long(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--long',
            object_fakes.container_name,
        ]
        verifylist = [
            ('long', True),
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name,
        )

        collist = ('Name', 'Bytes', 'Hash', 'Content Type', 'Last Modified')
        self.assertEqual(collist, columns)
        datalist = (
            (
                object_fakes.object_name_1,
                object_fakes.object_bytes_1,
                object_fakes.object_hash_1,
                object_fakes.object_content_type_1,
                object_fakes.object_modified_1,
            ),
            (
                object_fakes.object_name_2,
                object_fakes.object_bytes_2,
                object_fakes.object_hash_2,
                object_fakes.object_content_type_2,
                object_fakes.object_modified_2,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_object_list_objects_all(self):
        self.object_store_client.object_list.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--all',
            object_fakes.container_name,
        ]
        verifylist = [
            ('all', True),
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_list.assert_called_with(
            container=object_fakes.container_name,
            full_listing=True,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (object_fakes.object_name_1,),
            (object_fakes.object_name_2,),
        )
        self.assertEqual(datalist, tuple(data))


class TestObjectShow(object_fakes.TestObjectV1):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = obj.ShowObject(self.app, None)

    def test_object_show(self):
        self.object_store_client.object_show.return_value = copy.deepcopy(
            object_fakes.OBJECT
        )

        arglist = [
            object_fakes.container_name,
            object_fakes.object_name_1,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
            ('object', object_fakes.object_name_1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.object_store_client.object_show.assert_called_with(
            container=object_fakes.container_name,
            object=object_fakes.object_name_1,
        )

        collist = ('bytes', 'content_type', 'hash', 'last_modified', 'name')
        self.assertEqual(collist, columns)
        datalist = (
            object_fakes.object_bytes_1,
            object_fakes.object_content_type_1,
            object_fakes.object_hash_1,
            object_fakes.object_modified_1,
            object_fakes.object_name_1,
        )
        self.assertEqual(datalist, data)



#    Copyright 2017 FUJITSU LIMITED
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""VPN Utilities and helper functions."""

from typing import Any

from osc_lib import exceptions

from openstackclient.i18n import _

DPD_SUPPORTED_ACTIONS = [
    'hold',
    'clear',
    'restart',
    'restart-by-peer',
    'disabled',
]
DPD_SUPPORTED_KEYS = ['action', 'interval', 'timeout']

lifetime_keys = ['units', 'value']
lifetime_units = ['seconds']


def validate_dpd_dict(dpd_dict: dict[str, Any]) -> None:
    for key, value in dpd_dict.items():
        if key not in DPD_SUPPORTED_KEYS:
            message = _(
                "DPD Dictionary KeyError: "
                "Reason-Invalid DPD key : "
                "'%(key)s' not in %(supported_key)s"
            ) % {'key': key, 'supported_key': DPD_SUPPORTED_KEYS}
            raise exceptions.CommandError(message)
        if key == 'action' and value not in DPD_SUPPORTED_ACTIONS:
            message = _(
                "DPD Dictionary ValueError: "
                "Reason-Invalid DPD action : "
                "'%(key_value)s' not in %(supported_action)s"
            ) % {'key_value': value, 'supported_action': DPD_SUPPORTED_ACTIONS}
            raise exceptions.CommandError(message)
        if key in ('interval', 'timeout'):
            try:
                if int(value) <= 0:
                    raise ValueError()
            except ValueError:
                message = _(
                    "DPD Dictionary ValueError: "
                    "Reason-Invalid positive integer value: "
                    "'%(key)s' = %(value)s"
                ) % {'key': key, 'value': value}
                raise exceptions.CommandError(message)
            else:
                dpd_dict[key] = int(value)
    return


def validate_lifetime_dict(lifetime_dict: dict[str, Any]) -> None:

    for key, value in lifetime_dict.items():
        if key not in lifetime_keys:
            message = _(
                "Lifetime Dictionary KeyError: "
                "Reason-Invalid unit key : "
                "'%(key)s' not in %(supported_key)s"
            ) % {'key': key, 'supported_key': lifetime_keys}
            raise exceptions.CommandError(message)
        if key == 'units' and value not in lifetime_units:
            message = _(
                "Lifetime Dictionary ValueError: "
                "Reason-Invalid units : "
                "'%(key_value)s' not in %(supported_units)s"
            ) % {'key_value': key, 'supported_units': lifetime_units}
            raise exceptions.CommandError(message)
        if key == 'value':
            try:
                if int(value) < 60:
                    raise ValueError()
            except ValueError:
                message = _(
                    "Lifetime Dictionary ValueError: "
                    "Reason-Invalid value should be at least 60:"
                    "'%(key_value)s' = %(value)s"
                ) % {'key_value': key, 'value': value}
                raise exceptions.CommandError(message)
            else:
                lifetime_dict['value'] = int(value)
    return



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Security Groups Default Statefulness action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


def _get_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'name', 'revision_number']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateSecurityGroupDefaultStatefulness(command.ShowOne):
    _description = _(
        "Create a default statefulness setting for security groups"
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        stateful_group = parser.add_mutually_exclusive_group(required=True)
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            dest='stateful',
            help=_("Set default statefulness to stateful"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_false',
            default=None,
            dest='stateful',
            help=_("Set default statefulness to stateless"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "Apply the setting to this project (name or ID). "
                "If not specified, the setting applies system-wide"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs: dict[str, Any] = {}
        if parsed_args.stateful is not None:
            attrs['stateful'] = parsed_args.stateful
        project_id = None
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
        # Always set this, even if `None`, otherwise the Neutron API will
        # fulfill this value with the user project ID instead.
        attrs['project_id'] = project_id

        obj = client.create_security_groups_default_statefulness(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)


class DeleteSecurityGroupDefaultStatefulness(command.Command):
    _description = _("Delete security group default statefulness setting(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'setting',
            metavar='<setting>',
            nargs="+",
            help=_("Default statefulness setting(s) to delete (ID only)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        result = 0
        client = self.app.client_manager.network
        for s in parsed_args.setting:
            try:
                obj = client.find_security_groups_default_statefulness(
                    s, ignore_missing=False
                )
                client.delete_security_groups_default_statefulness(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete default statefulness setting "
                        "with ID '%(id)s': %(e)s"
                    ),
                    {'id': s, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.setting)
            msg = _(
                "%(result)s of %(total)s default statefulness settings "
                "failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListSecurityGroupDefaultStatefulness(command.Lister):
    _description = _("List security group default statefulness settings")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("List only settings for this project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        column_headers = (
            'ID',
            'Project ID',
            'Stateful',
        )
        columns = (
            'id',
            'project_id',
            'stateful',
        )

        query: dict[str, Any] = {}
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            query['project_id'] = project_id

        data = client.security_groups_default_statefulness(**query)

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class SetSecurityGroupDefaultStatefulness(command.Command):
    _description = _("Update a security group default statefulness setting")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'setting',
            metavar='<setting>',
            help=_("Default statefulness setting to modify (ID only)"),
        )
        stateful_group = parser.add_mutually_exclusive_group(required=True)
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            dest='stateful',
            help=_("Set default statefulness to stateful"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_false',
            default=None,
            dest='stateful',
            help=_("Set default statefulness to stateless"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_security_groups_default_statefulness(
            parsed_args.setting, ignore_missing=False
        )
        attrs: dict[str, Any] = {}
        if parsed_args.stateful is not None:
            attrs['stateful'] = parsed_args.stateful
        client.update_security_groups_default_statefulness(obj, **attrs)


class ShowSecurityGroupDefaultStatefulness(command.ShowOne):
    _description = _("Show a security group default statefulness setting")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'setting',
            metavar='<setting>',
            help=_("Default statefulness setting to display (ID only)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_security_groups_default_statefulness(
            parsed_args.setting, ignore_missing=False
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call
import uuid

from openstack.network.v2 import (
    default_security_group_rule as _default_security_group_rule,
)
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network import utils as network_utils
from openstackclient.network.v2 import default_security_group_rule
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestCreateDefaultSecurityGroupRule(network_fakes.TestNetworkV2):
    expected_columns = (
        'description',
        'direction',
        'ether_type',
        'id',
        'port_range_max',
        'port_range_min',
        'protocol',
        'remote_address_group_id',
        'remote_group_id',
        'remote_ip_prefix',
        'used_in_default_sg',
        'used_in_non_default_sg',
    )

    expected_data = None

    def _setup_default_security_group_rule(self, attrs=None):
        default_security_group_rule_attrs = {
            'description': 'default-security-group-rule-description-'
            + uuid.uuid4().hex,
            'direction': 'ingress',
            'ether_type': 'IPv4',
            'id': 'default-security-group-rule-id-' + uuid.uuid4().hex,
            'port_range_max': None,
            'port_range_min': None,
            'protocol': None,
            'remote_group_id': None,
            'remote_address_group_id': None,
            'remote_ip_prefix': '0.0.0.0/0',
            'location': 'MUNCHMUNCHMUNCH',
            'used_in_default_sg': False,
            'used_in_non_default_sg': False,
        }
        attrs = attrs or {}
        # Overwrite default attributes.
        default_security_group_rule_attrs.update(attrs)
        self._default_sg_rule = sdk_fakes.generate_fake_resource(
            _default_security_group_rule.DefaultSecurityGroupRule,
            **default_security_group_rule_attrs,
        )

        self.network_client.create_default_security_group_rule.return_value = (
            self._default_sg_rule
        )
        self.expected_data = (
            self._default_sg_rule.description,
            self._default_sg_rule.direction,
            self._default_sg_rule.ether_type,
            self._default_sg_rule.id,
            self._default_sg_rule.port_range_max,
            self._default_sg_rule.port_range_min,
            self._default_sg_rule.protocol,
            self._default_sg_rule.remote_address_group_id,
            self._default_sg_rule.remote_group_id,
            self._default_sg_rule.remote_ip_prefix,
            self._default_sg_rule.used_in_default_sg,
            self._default_sg_rule.used_in_non_default_sg,
        )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = default_security_group_rule.CreateDefaultSecurityGroupRule(
            self.app, None
        )

    def test_create_all_remote_options(self):
        arglist = [
            '--remote-ip',
            '10.10.0.0/24',
            '--remote-group',
            'test-remote-group-id',
            '--remote-address-group',
            'test-remote-address-group-id',
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            [],
        )

    def test_create_bad_ethertype(self):
        arglist = [
            '--ethertype',
            'foo',
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            [],
        )

    def test_lowercase_ethertype(self):
        arglist = [
            '--ethertype',
            'ipv4',
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv4', parsed_args.ethertype)

    def test_lowercase_v6_ethertype(self):
        arglist = [
            '--ethertype',
            'ipv6',
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv6', parsed_args.ethertype)

    def test_proper_case_ethertype(self):
        arglist = [
            '--ethertype',
            'IPv6',
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv6', parsed_args.ethertype)

    def test_create_all_port_range_options(self):
        arglist = [
            '--dst-port',
            '80:80',
            '--icmp-type',
            '3',
            '--icmp-code',
            '1',
        ]
        verifylist = [
            ('dst_port', (80, 80)),
            ('icmp_type', 3),
            ('icmp_code', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_default_rule(self):
        self._setup_default_security_group_rule(
            {
                'protocol': 'tcp',
                'port_range_max': 443,
                'port_range_min': 443,
            }
        )
        arglist = [
            '--protocol',
            'tcp',
            '--dst-port',
            str(self._default_sg_rule.port_range_min),
        ]
        verifylist = [
            (
                'dst_port',
                (
                    self._default_sg_rule.port_range_min,
                    self._default_sg_rule.port_range_max,
                ),
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_max': self._default_sg_rule.port_range_max,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def _test_create_protocol_any_helper(
        self, for_default_sg=False, for_custom_sg=False
    ):
        self._setup_default_security_group_rule(
            {
                'protocol': None,
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            'any',
            '--remote-ip',
            self._default_sg_rule.remote_ip_prefix,
        ]
        if for_default_sg:
            arglist.append('--for-default-sg')
        if for_custom_sg:
            arglist.append('--for-custom-sg')
        verifylist = [
            ('protocol', 'any'),
            ('remote_ip', self._default_sg_rule.remote_ip_prefix),
            ('for_default_sg', for_default_sg),
            ('for_custom_sg', for_custom_sg),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': for_default_sg,
                'used_in_non_default_sg': for_custom_sg,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_protocol_any_not_for_default_sg(self):
        self._test_create_protocol_any_helper()

    def test_create_protocol_any_for_default_sg(self):
        self._test_create_protocol_any_helper(for_default_sg=True)

    def test_create_protocol_any_for_custom_sg(self):
        self._test_create_protocol_any_helper(for_custom_sg=True)

    def test_create_protocol_any_for_default_and_custom_sg(self):
        self._test_create_protocol_any_helper(
            for_default_sg=True, for_custom_sg=True
        )

    def test_create_remote_address_group(self):
        self._setup_default_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_address_group_id': 'remote-address-group-id',
            }
        )
        arglist = [
            '--protocol',
            'icmp',
            '--remote-address-group',
            self._default_sg_rule.remote_address_group_id,
        ]
        verifylist = [
            (
                'remote_address_group',
                self._default_sg_rule.remote_address_group_id,
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_address_group_id': self._default_sg_rule.remote_address_group_id,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_remote_group(self):
        self._setup_default_security_group_rule(
            {
                'protocol': 'tcp',
                'port_range_max': 22,
                'port_range_min': 22,
            }
        )
        arglist = [
            '--protocol',
            'tcp',
            '--dst-port',
            str(self._default_sg_rule.port_range_min),
            '--ingress',
            '--remote-group',
            'remote-group-id',
        ]
        verifylist = [
            (
                'dst_port',
                (
                    self._default_sg_rule.port_range_min,
                    self._default_sg_rule.port_range_max,
                ),
            ),
            ('ingress', True),
            ('remote_group', 'remote-group-id'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_max': self._default_sg_rule.port_range_max,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_group_id': 'remote-group-id',
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_source_group(self):
        self._setup_default_security_group_rule(
            {
                'remote_group_id': 'remote-group-id',
            }
        )
        arglist = [
            '--ingress',
            '--remote-group',
            'remote-group-id',
        ]
        verifylist = [
            ('ingress', True),
            ('remote_group', 'remote-group-id'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_group_id': 'remote-group-id',
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_source_ip(self):
        self._setup_default_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            self._default_sg_rule.protocol,
            '--remote-ip',
            self._default_sg_rule.remote_ip_prefix,
        ]
        verifylist = [
            ('protocol', self._default_sg_rule.protocol),
            ('remote_ip', self._default_sg_rule.remote_ip_prefix),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_remote_ip(self):
        self._setup_default_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            self._default_sg_rule.protocol,
            '--remote-ip',
            self._default_sg_rule.remote_ip_prefix,
        ]
        verifylist = [
            ('protocol', self._default_sg_rule.protocol),
            ('remote_ip', self._default_sg_rule.remote_ip_prefix),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_tcp_with_icmp_type(self):
        arglist = [
            '--protocol',
            'tcp',
            '--icmp-type',
            '15',
        ]
        verifylist = [
            ('protocol', 'tcp'),
            ('icmp_type', 15),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_icmp_code(self):
        arglist = [
            '--protocol',
            '1',
            '--icmp-code',
            '1',
        ]
        verifylist = [
            ('protocol', '1'),
            ('icmp_code', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_icmp_code_zero(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': 0,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._default_sg_rule.protocol,
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--icmp-code',
            str(self._default_sg_rule.port_range_max),
        ]
        verifylist = [
            ('protocol', self._default_sg_rule.protocol),
            ('icmp_code', self._default_sg_rule.port_range_max),
            ('icmp_type', self._default_sg_rule.port_range_min),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_code_greater_than_zero(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': 18,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._default_sg_rule.protocol,
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--icmp-code',
            str(self._default_sg_rule.port_range_max),
        ]
        verifylist = [
            ('protocol', self._default_sg_rule.protocol),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', self._default_sg_rule.port_range_max),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_code_negative_value(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': None,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._default_sg_rule.protocol,
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--icmp-code',
            '-2',
        ]
        verifylist = [
            ('protocol', self._default_sg_rule.protocol),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', -2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 15,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_zero(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 0,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_greater_than_zero(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': 13,  # timestamp
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_negative_value(self):
        self._setup_default_security_group_rule(
            {
                'port_range_min': None,  # timestamp
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            '-13',
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', -13),
            ('icmp_code', None),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_ipv6_icmp_type_code(self):
        self._setup_default_security_group_rule(
            {
                'ether_type': 'IPv6',
                'port_range_min': 139,
                'port_range_max': 2,
                'protocol': 'ipv6-icmp',
                'remote_ip_prefix': '::/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--icmp-code',
            str(self._default_sg_rule.port_range_max),
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', self._default_sg_rule.port_range_max),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_min': self._default_sg_rule.port_range_min,
                'port_range_max': self._default_sg_rule.port_range_max,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmpv6_type(self):
        self._setup_default_security_group_rule(
            {
                'ether_type': 'IPv6',
                'port_range_min': 139,
                'protocol': 'icmpv6',
                'remote_ip_prefix': '::/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._default_sg_rule.port_range_min),
            '--protocol',
            self._default_sg_rule.protocol,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._default_sg_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._default_sg_rule.protocol),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'port_range_min': self._default_sg_rule.port_range_min,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_with_description(self):
        self._setup_default_security_group_rule(
            {
                'description': 'Setting SGR',
            }
        )
        arglist = [
            '--description',
            self._default_sg_rule.description,
        ]
        verifylist = [
            ('description', self._default_sg_rule.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_default_security_group_rule.assert_called_once_with(
            **{
                'description': self._default_sg_rule.description,
                'direction': self._default_sg_rule.direction,
                'ethertype': self._default_sg_rule.ether_type,
                'protocol': self._default_sg_rule.protocol,
                'remote_ip_prefix': self._default_sg_rule.remote_ip_prefix,
                'used_in_default_sg': False,
                'used_in_non_default_sg': False,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)


class TestDeleteDefaultSecurityGroupRule(network_fakes.TestNetworkV2):
    # The default security group rules to be deleted.
    default_security_group_rule_attrs = {
        'direction': 'ingress',
        'ether_type': 'IPv4',
        'port_range_max': None,
        'port_range_min': None,
        'protocol': None,
        'remote_group_id': None,
        'remote_address_group_id': None,
        'remote_ip_prefix': '0.0.0.0/0',
        'location': 'MUNCHMUNCHMUNCH',
        'used_in_default_sg': False,
        'used_in_non_default_sg': False,
    }
    _default_sg_rules = list(
        sdk_fakes.generate_fake_resources(
            _default_security_group_rule.DefaultSecurityGroupRule,
            count=2,
            attrs=default_security_group_rule_attrs,
        )
    )

    def setUp(self):
        super().setUp()

        self.network_client.delete_default_security_group_rule.return_value = (
            None
        )

        # Get the command object to test
        self.cmd = default_security_group_rule.DeleteDefaultSecurityGroupRule(
            self.app, None
        )

    def test_default_security_group_rule_delete(self):
        arglist = [
            self._default_sg_rules[0].id,
        ]
        verifylist = [
            ('rule', [self._default_sg_rules[0].id]),
        ]
        self.network_client.find_default_security_group_rule.return_value = (
            self._default_sg_rules[0]
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_default_security_group_rule.assert_called_once_with(
            self._default_sg_rules[0]
        )
        self.assertIsNone(result)

    def test_multi_default_security_group_rules_delete(self):
        arglist = []
        verifylist = []

        for s in self._default_sg_rules:
            arglist.append(s.id)
        verifylist = [
            ('rule', arglist),
        ]
        self.network_client.find_default_security_group_rule.side_effect = (
            self._default_sg_rules
        )
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for s in self._default_sg_rules:
            calls.append(call(s))
        self.network_client.delete_default_security_group_rule.assert_has_calls(
            calls
        )
        self.assertIsNone(result)

    def test_multi_default_security_group_rules_delete_with_exception(self):
        arglist = [
            self._default_sg_rules[0].id,
            'unexist_rule',
        ]
        verifylist = [
            ('rule', [self._default_sg_rules[0].id, 'unexist_rule']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [
            self._default_sg_rules[0],
            exceptions.CommandError,
        ]
        self.network_client.find_default_security_group_rule.side_effect = (
            find_mock_result
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 default rules failed to delete.', str(e))

        self.network_client.find_default_security_group_rule.assert_any_call(
            self._default_sg_rules[0].id, ignore_missing=False
        )
        self.network_client.find_default_security_group_rule.assert_any_call(
            'unexist_rule', ignore_missing=False
        )
        self.network_client.delete_default_security_group_rule.assert_called_once_with(
            self._default_sg_rules[0]
        )


class TestListDefaultSecurityGroupRule(network_fakes.TestNetworkV2):
    # The security group rule to be listed.
    _default_sg_rule_tcp = sdk_fakes.generate_fake_resource(
        _default_security_group_rule.DefaultSecurityGroupRule,
        **{'protocol': 'tcp', 'port_range_max': 80, 'port_range_min': 80},
    )
    _default_sg_rule_icmp = sdk_fakes.generate_fake_resource(
        _default_security_group_rule.DefaultSecurityGroupRule,
        **{'protocol': 'icmp', 'remote_ip_prefix': '10.0.2.0/24'},
    )
    _default_sg_rules = [
        _default_sg_rule_tcp,
        _default_sg_rule_icmp,
    ]

    expected_columns = (
        'ID',
        'IP Protocol',
        'Ethertype',
        'IP Range',
        'Port Range',
        'Direction',
        'Remote Security Group',
        'Remote Address Group',
        'Used in default Security Group',
        'Used in custom Security Group',
    )

    expected_data = []
    expected_data_no_group = []
    for _default_sg_rule in _default_sg_rules:
        expected_data.append(
            (
                _default_sg_rule.id,
                _default_sg_rule.protocol,
                _default_sg_rule.ether_type,
                _default_sg_rule.remote_ip_prefix,
                network_utils.format_network_port_range(_default_sg_rule),
                _default_sg_rule.direction,
                _default_sg_rule.remote_group_id,
                _default_sg_rule.remote_address_group_id,
                _default_sg_rule.used_in_default_sg,
                _default_sg_rule.used_in_non_default_sg,
            )
        )

    def setUp(self):
        super().setUp()

        self.network_client.default_security_group_rules.return_value = (
            self._default_sg_rules
        )

        # Get the command object to test
        self.cmd = default_security_group_rule.ListDefaultSecurityGroupRule(
            self.app, None
        )

    def test_list_default(self):
        self._default_sg_rule_tcp.port_range_min = 80
        parsed_args = self.check_parser(self.cmd, [], [])

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.default_security_group_rules.assert_called_once_with(
            **{}
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, list(data))

    def test_list_with_protocol(self):
        self._default_sg_rule_tcp.port_range_min = 80
        arglist = [
            '--protocol',
            'tcp',
        ]
        verifylist = [
            ('protocol', 'tcp'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.default_security_group_rules.assert_called_once_with(
            **{
                'protocol': 'tcp',
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, list(data))

    def test_list_with_ingress(self):
        self._default_sg_rule_tcp.port_range_min = 80
        arglist = [
            '--ingress',
        ]
        verifylist = [
            ('ingress', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.default_security_group_rules.assert_called_once_with(
            **{
                'direction': 'ingress',
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, list(data))

    def test_list_with_wrong_egress(self):
        self._default_sg_rule_tcp.port_range_min = 80
        arglist = [
            '--egress',
        ]
        verifylist = [
            ('egress', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.default_security_group_rules.assert_called_once_with(
            **{
                'direction': 'egress',
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, list(data))


class TestShowDefaultSecurityGroupRule(network_fakes.TestNetworkV2):
    # The default security group rule to be shown.
    _default_sg_rule = sdk_fakes.generate_fake_resource(
        _default_security_group_rule.DefaultSecurityGroupRule
    )

    columns = (
        'description',
        'direction',
        'ether_type',
        'id',
        'port_range_max',
        'port_range_min',
        'protocol',
        'remote_address_group_id',
        'remote_group_id',
        'remote_ip_prefix',
        'used_in_default_sg',
        'used_in_non_default_sg',
    )

    data = (
        _default_sg_rule.description,
        _default_sg_rule.direction,
        _default_sg_rule.ether_type,
        _default_sg_rule.id,
        _default_sg_rule.port_range_max,
        _default_sg_rule.port_range_min,
        _default_sg_rule.protocol,
        _default_sg_rule.remote_address_group_id,
        _default_sg_rule.remote_group_id,
        _default_sg_rule.remote_ip_prefix,
        _default_sg_rule.used_in_default_sg,
        _default_sg_rule.used_in_non_default_sg,
    )

    def setUp(self):
        super().setUp()

        self.network_client.find_default_security_group_rule.return_value = (
            self._default_sg_rule
        )

        # Get the command object to test
        self.cmd = default_security_group_rule.ShowDefaultSecurityGroupRule(
            self.app, None
        )

    def test_show_no_options(self):
        self.assertRaises(
            tests_utils.ParserException, self.check_parser, self.cmd, [], []
        )

    def test_show_all_options(self):
        arglist = [
            self._default_sg_rule.id,
        ]
        verifylist = [
            ('rule', self._default_sg_rule.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_default_security_group_rule.assert_called_once_with(
            self._default_sg_rule.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



# All Rights Reserved 2020
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import operator
import uuid

from openstack.network.v2 import tap_service
from openstack.test import fakes as sdk_fakes
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient.network.v2.taas import tap_service as osc_tap_service
from openstackclient.tests.unit.network.v2 import fakes as network_fakes


columns_long = tuple(
    col
    for col, _, listing_mode in osc_tap_service._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
headers_long = tuple(
    head
    for _, head, listing_mode in osc_tap_service._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
sorted_attr_map = sorted(osc_tap_service._attr_map, key=operator.itemgetter(1))
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(attrs, columns)


class TestCreateTapService(network_fakes.TestNetworkV2):
    columns = (
        'description',
        'id',
        'name',
        'port_id',
        'project_id',
        'status',
    )

    def setUp(self):
        super().setUp()
        self.cmd = osc_tap_service.CreateTapService(self.app, None)

    def test_create_tap_service(self):
        """Test Create Tap Service."""
        port_id = str(uuid.uuid4())
        fake_port = network_fakes.create_one_port(attrs={'id': port_id})
        fake_tap_service = sdk_fakes.generate_fake_resource(
            tap_service.TapService, **{'port_id': port_id}
        )
        self.network_client.create_tap_service.return_value = fake_tap_service
        self.network_client.find_port.return_value = fake_port
        self.network_client.find_tap_service.side_effect = (
            lambda _, name_or_id: {'id': name_or_id}
        )
        arg_list = [
            '--name',
            fake_tap_service['name'],
            '--port',
            fake_tap_service['port_id'],
        ]

        verify_list = [
            ('name', fake_tap_service['name']),
            ('port_id', fake_tap_service['port_id']),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        self.network_client.find_tap_service.return_value = fake_tap_service

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_tap_service.assert_called_once_with(
            **{
                'name': fake_tap_service['name'],
                'port_id': fake_tap_service['port_id'],
            }
        )
        self.assertEqual(self.columns, columns)
        fake_data = _get_data(
            fake_tap_service, osc_tap_service._get_columns(fake_tap_service)[1]
        )
        self.assertEqual(fake_data, data)


class TestListTapService(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.cmd = osc_tap_service.ListTapService(self.app, None)

    def test_list_tap_service(self):
        """Test List Tap Service."""
        fake_tap_services = list(
            sdk_fakes.generate_fake_resources(tap_service.TapService, count=4)
        )
        self.network_client.tap_services.return_value = fake_tap_services

        arg_list = []
        verify_list = []

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.tap_services.assert_called_once()
        self.assertEqual(headers, list(headers_long))
        self.assertCountEqual(
            list(data),
            [
                _get_data(fake_tap_service, columns_long)
                for fake_tap_service in fake_tap_services
            ],
        )


class TestDeleteTapService(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.network_client.find_tap_service.side_effect = (
            lambda name_or_id, ignore_missing: tap_service.TapService(
                id=name_or_id
            )
        )
        self.cmd = osc_tap_service.DeleteTapService(self.app, None)

    def test_delete_tap_service(self):
        """Test Delete tap service."""

        fake_tap_service = sdk_fakes.generate_fake_resource(
            tap_service.TapService
        )

        arg_list = [
            fake_tap_service['id'],
        ]
        verify_list = [
            (osc_tap_service.TAP_SERVICE, [fake_tap_service['id']]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_tap_service.assert_called_once_with(
            fake_tap_service['id']
        )
        self.assertIsNone(result)


class TestShowTapService(network_fakes.TestNetworkV2):
    columns = (
        'description',
        'id',
        'name',
        'port_id',
        'project_id',
        'status',
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_tap_service.side_effect = (
            lambda name_or_id, ignore_missing: tap_service.TapService(
                id=name_or_id
            )
        )
        self.cmd = osc_tap_service.ShowTapService(self.app, None)

    def test_show_tap_service(self):
        """Test Show tap service."""

        fake_tap_service = sdk_fakes.generate_fake_resource(
            tap_service.TapService
        )
        self.network_client.get_tap_service.return_value = fake_tap_service
        arg_list = [
            fake_tap_service['id'],
        ]
        verify_list = [
            (osc_tap_service.TAP_SERVICE, fake_tap_service['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.get_tap_service.assert_called_once_with(
            fake_tap_service['id']
        )
        self.assertEqual(self.columns, headers)
        fake_data = _get_data(
            fake_tap_service, osc_tap_service._get_columns(fake_tap_service)[1]
        )
        self.assertEqual(fake_data, data)


class TestUpdateTapService(network_fakes.TestNetworkV2):
    _new_name = 'new_name'

    columns = (
        'description',
        'id',
        'name',
        'port_id',
        'project_id',
        'status',
    )

    def setUp(self):
        super().setUp()
        self.cmd = osc_tap_service.UpdateTapService(self.app, None)
        self.network_client.find_tap_service.side_effect = (
            lambda name_or_id, ignore_missing: tap_service.TapService(
                id=name_or_id
            )
        )

    def test_update_tap_service(self):
        """Test update tap service"""
        fake_tap_service = sdk_fakes.generate_fake_resource(
            tap_service.TapService
        )
        new_tap_service = copy.deepcopy(fake_tap_service)
        new_tap_service['name'] = self._new_name

        self.network_client.update_tap_service.return_value = new_tap_service

        arg_list = [
            fake_tap_service['id'],
            '--name',
            self._new_name,
        ]
        verify_list = [('name', self._new_name)]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {'name': self._new_name}

        self.network_client.update_tap_service.assert_called_once_with(
            fake_tap_service['id'], **attrs
        )
        self.assertEqual(self.columns, columns)
        fake_data = _get_data(
            new_tap_service, osc_tap_service._get_columns(new_tap_service)[1]
        )
        self.assertEqual(fake_data, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import ddt
from openstack.block_storage.v3 import service as _service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import block_storage_log_level as service


class TestBlockStorageLogLevelList(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.log_level = sdk_fakes.generate_fake_resource(
            _service.LogLevel, binary='cinder-scheduler'
        )
        self.volume_client.get_service_log_levels.return_value = [
            self.log_level
        ]

        self.cmd = service.BlockStorageLogLevelList(self.app, None)

    def test_block_storage_log_level_list(self):
        self.set_volume_api_version('3.32')

        arglist = [
            '--host',
            self.log_level.host,
            '--service',
            self.log_level.binary,
            '--log-prefix',
            'cinder.',
        ]
        verifylist = [
            ('host', self.log_level.host),
            ('service', self.log_level.binary),
            ('log_prefix', 'cinder.'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Binary',
            'Host',
            'Prefix',
            'Level',
        ]
        datalist = tuple(
            (
                self.log_level.binary,
                self.log_level.host,
                prefix,
                level,
            )
            for prefix, level in self.log_level.levels.values()
        )
        self.assertEqual(expected_columns, columns)
        self.assertEqual(datalist, tuple(data))

        self.volume_client.get_service_log_levels.assert_called_with(
            server=self.log_level.host,
            binary=self.log_level.binary,
            prefix='cinder.',
        )

    def test_block_storage_log_level_list_pre_332(self):
        arglist = [
            '--host',
            self.log_level.host,
            '--service',
            'cinder-api',
            '--log-prefix',
            'cinder_test.api.common',
        ]
        verifylist = [
            ('host', self.log_level.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder_test.api.common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.32 or greater is required', str(exc)
        )

    def test_block_storage_log_level_list_invalid_service_name(self):
        self.set_volume_api_version('3.32')

        arglist = [
            '--host',
            self.log_level.host,
            '--service',
            'nova-api',
            '--log-prefix',
            'cinder_test.api.common',
        ]
        verifylist = [
            ('host', self.log_level.host),
            ('service', 'nova-api'),
            ('log_prefix', 'cinder_test.api.common'),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


@ddt.ddt
class TestBlockStorageLogLevelSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.log_level = sdk_fakes.generate_fake_resource(
            _service.LogLevel, binary='cinder-api'
        )
        self.volume_client.set_service_log_levels.return_value = None

        self.cmd = service.BlockStorageLogLevelSet(self.app, None)

    def test_block_storage_log_level_set(self):
        self.set_volume_api_version('3.32')

        arglist = [
            'ERROR',
            '--host',
            self.log_level.host,
            '--service',
            self.log_level.binary,
            '--log-prefix',
            'cinder.api.common',
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.log_level.host),
            ('service', self.log_level.binary),
            ('log_prefix', 'cinder.api.common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ret = self.cmd.take_action(parsed_args)

        self.assertIsNone(ret)
        self.volume_client.set_service_log_levels.assert_called_with(
            level='ERROR',
            server=self.log_level.host,
            binary=self.log_level.binary,
            prefix='cinder.api.common',
        )

    def test_block_storage_log_level_set_pre_332(self):
        arglist = [
            'ERROR',
            '--host',
            self.log_level.host,
            '--service',
            'cinder-api',
            '--log-prefix',
            'cinder.api.common',
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.log_level.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder.api.common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.32 or greater is required', str(exc)
        )

    def test_block_storage_log_level_set_invalid_service_name(self):
        self.set_volume_api_version('3.32')

        arglist = [
            'ERROR',
            '--host',
            self.log_level.host,
            '--service',
            'nova-api',
            '--log-prefix',
            'cinder.api.common',
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.log_level.host),
            ('service', 'nova-api'),
            ('log_prefix', 'cinder.api.common'),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    @ddt.data('WARNING', 'info', 'Error', 'debuG', 'fake-log-level')
    def test_block_storage_log_level_set_log_level(self, log_level):
        self.set_volume_api_version('3.32')

        arglist = [
            log_level,
            '--host',
            self.log_level.host,
            '--service',
            'cinder-api',
            '--log-prefix',
            'cinder.api.common',
        ]
        verifylist = [
            ('level', log_level.upper()),
            ('host', self.log_level.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder.api.common'),
        ]

        if log_level == 'fake-log-level':
            self.assertRaises(
                tests_utils.ParserException,
                self.check_parser,
                self.cmd,
                arglist,
                verifylist,
            )
        else:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)

            self.cmd.take_action(parsed_args)

            self.volume_client.set_service_log_levels.assert_called_with(
                level=log_level.upper(),
                server=self.log_level.host,
                binary=self.log_level.binary,
                prefix='cinder.api.common',
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import time
from typing import cast
import uuid

from tempest.lib import exceptions

from openstackclient.tests.functional import base


class ComputeTestCase(base.TestCase):
    """Common functional test bits for Compute commands"""

    flavor_name: str
    image_name: str
    network_arg: str

    def setUp(self):
        """Select common resources"""
        super().setUp()
        self.flavor_name = self.get_flavor()
        self.image_name = self.get_image()
        self.network_arg = self.get_network()

    @classmethod
    def get_flavor(cls) -> str:
        valid_flavors = ['m1.tiny', 'cirros256']
        # NOTE(rtheis): Get cirros256 or m1.tiny flavors since functional
        #               tests may create other flavors.
        flavors = cls.openstack("flavor list", parse_output=True)
        for flavor in flavors:
            if flavor['Name'] in ['m1.tiny', 'cirros256']:
                return cast(str, flavor['Name'])

        raise Exception(
            f'Failed to find a suitable flavor. Required one of: '
            f'{", ".join(valid_flavors)}'
        )

    @classmethod
    def get_image(cls) -> str:
        # NOTE(rtheis): Get first Cirros image since functional tests may
        #               create other images.  Image may be named '-uec' or
        #               '-disk'.
        images = cls.openstack("image list", parse_output=True)
        for image in images:
            if image['Name'].startswith('cirros-') and (
                image['Name'].endswith('-uec')
                or image['Name'].endswith('-disk')
            ):
                return cast(str, image['Name'])

        raise Exception(
            'Failed to find a suitable image. Required one matching one of '
            'the following patterns: cirros-*-uec, cirros-*-disk'
        )

    @classmethod
    def get_network(cls) -> str:
        try:
            # NOTE(rtheis): Get private network since functional tests may
            #               create other networks.
            cmd_output = cls.openstack(
                'network show private',
                parse_output=True,
            )
        except exceptions.CommandFailed:
            return ''

        return '='.join(('--nic net-id', cmd_output['id']))

    def server_create(self, name=None, cleanup=True):
        """Create server, with cleanup"""
        if not self.flavor_name:
            self.flavor_name = self.get_flavor()
        if not self.image_name:
            self.image_name = self.get_image()
        if not self.network_arg:
            self.network_arg = self.get_network()
        name = name or uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + self.network_arg
            + ' '
            + '--wait '
            + name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            name,
            cmd_output["name"],
        )
        if cleanup:
            self.addCleanup(self.server_delete, name)
        return cmd_output

    def server_delete(self, name):
        """Delete server by name"""
        raw_output = self.openstack('server delete ' + name)
        self.assertOutput('', raw_output)

    def wait_for_status(
        self,
        name,
        expected_status='ACTIVE',
        wait=900,
        interval=10,
    ):
        """Wait until server reaches expected status"""
        # TODO(thowe): Add a server wait command to osc
        failures = ['ERROR']
        total_sleep = 0
        while total_sleep < wait:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            status = cmd_output['status']
            if status == expected_status:
                print(f'Server {name} now has status {status}')
                break
            print(
                f'Server {name}: Waiting for {expected_status}, current status: {status}'
            )
            self.assertNotIn(status, failures)
            time.sleep(interval)
            total_sleep += interval

        cmd_output = self.openstack(
            'server show ' + name,
            parse_output=True,
        )
        status = cmd_output['status']
        self.assertEqual(status, expected_status)
        # give it a little bit more time
        time.sleep(5)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

import ddt
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2 import network_rbac
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestNetworkRBAC(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()


@ddt.ddt
class TestCreateNetworkRBAC(TestNetworkRBAC):
    network_object = network_fakes.create_one_network()
    qos_object = network_fakes.create_one_qos_policy()
    sg_object = network_fakes.create_one_security_group()
    as_object = network_fakes.create_one_address_scope()
    snp_object = network_fakes.FakeSubnetPool.create_one_subnet_pool()
    ag_object = network_fakes.create_one_address_group()
    project = sdk_fakes.generate_fake_resource(_project.Project)
    rbac_policy = network_fakes.create_one_network_rbac(
        attrs={
            'project_id': project.id,
            'target_tenant': project.id,
            'object_id': network_object.id,
        }
    )

    columns = (
        'action',
        'id',
        'object_id',
        'object_type',
        'project_id',
        'target_project_id',
    )

    data = [
        rbac_policy.action,
        rbac_policy.id,
        rbac_policy.object_id,
        rbac_policy.object_type,
        rbac_policy.project_id,
        rbac_policy.target_project_id,
    ]

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = network_rbac.CreateNetworkRBAC(self.app, None)

        self.network_client.create_rbac_policy.return_value = self.rbac_policy

        self.network_client.find_network.return_value = self.network_object

        self.network_client.find_qos_policy.return_value = self.qos_object

        self.network_client.find_security_group.return_value = self.sg_object

        self.network_client.find_address_scope.return_value = self.as_object

        self.network_client.find_subnet_pool.return_value = self.snp_object

        self.network_client.find_address_group.return_value = self.ag_object

        self.identity_sdk_client.find_project.return_value = self.project

    def test_network_rbac_create_no_type(self):
        arglist = [
            '--action',
            self.rbac_policy.action,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('action', self.rbac_policy.action),
            ('rbac_policy', self.rbac_policy.id),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_rbac_create_no_action(self):
        arglist = [
            '--type',
            self.rbac_policy.object_type,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('type', self.rbac_policy.object_type),
            ('rbac_policy', self.rbac_policy.id),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_rbac_create_invalid_type(self):
        arglist = [
            '--action',
            self.rbac_policy.action,
            '--type',
            'invalid_type',
            '--target-project',
            self.rbac_policy.target_project_id,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('action', self.rbac_policy.action),
            ('type', 'invalid_type'),
            ('target-project', self.rbac_policy.target_project_id),
            ('rbac_policy', self.rbac_policy.id),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_rbac_create_invalid_action(self):
        arglist = [
            '--type',
            self.rbac_policy.object_type,
            '--action',
            'invalid_action',
            '--target-project',
            self.rbac_policy.target_project_id,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('type', self.rbac_policy.object_type),
            ('action', 'invalid_action'),
            ('target-project', self.rbac_policy.target_project_id),
            ('rbac_policy', self.rbac_policy.id),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_rbac_create(self):
        arglist = [
            '--type',
            self.rbac_policy.object_type,
            '--action',
            self.rbac_policy.action,
            '--target-project',
            self.rbac_policy.target_project_id,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('type', self.rbac_policy.object_type),
            ('action', self.rbac_policy.action),
            ('target_project', self.rbac_policy.target_project_id),
            ('rbac_object', self.rbac_policy.object_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_rbac_policy.assert_called_with(
            **{
                'object_id': self.rbac_policy.object_id,
                'object_type': self.rbac_policy.object_type,
                'action': self.rbac_policy.action,
                'target_tenant': self.rbac_policy.target_project_id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_rbac_create_with_target_all_projects(self):
        arglist = [
            '--type',
            self.rbac_policy.object_type,
            '--action',
            self.rbac_policy.action,
            '--target-all-projects',
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('type', self.rbac_policy.object_type),
            ('action', self.rbac_policy.action),
            ('target_all_projects', True),
            ('rbac_object', self.rbac_policy.object_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        _columns, _data = self.cmd.take_action(parsed_args)

        self.network_client.create_rbac_policy.assert_called_with(
            **{
                'object_id': self.rbac_policy.object_id,
                'object_type': self.rbac_policy.object_type,
                'action': self.rbac_policy.action,
                'target_tenant': '*',
            }
        )

    def test_network_rbac_create_all_options(self):
        arglist = [
            '--type',
            self.rbac_policy.object_type,
            '--action',
            self.rbac_policy.action,
            '--target-project',
            self.rbac_policy.target_project_id,
            '--project',
            self.rbac_policy.project_id,
            '--project-domain',
            self.project.domain_id,
            '--target-project-domain',
            self.project.domain_id,
            self.rbac_policy.object_id,
        ]
        verifylist = [
            ('type', self.rbac_policy.object_type),
            ('action', self.rbac_policy.action),
            ('target_project', self.rbac_policy.target_project_id),
            ('project', self.rbac_policy.project_id),
            ('project_domain', self.project.domain_id),
            ('target_project_domain', self.project.domain_id),
            ('rbac_object', self.rbac_policy.object_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_rbac_policy.assert_called_with(
            **{
                'object_id': self.rbac_policy.object_id,
                'object_type': self.rbac_policy.object_type,
                'action': self.rbac_policy.action,
                'target_tenant': self.rbac_policy.target_project_id,
                'project_id': self.rbac_policy.project_id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    @ddt.data(
        ('qos_policy', "qos_object"),
        ('security_group', "sg_object"),
        ('subnetpool', "snp_object"),
        ('address_scope', "as_object"),
        ('address_group', "ag_object"),
    )
    @ddt.unpack
    def test_network_rbac_create_object(self, obj_type, obj_fake_attr):
        obj_fake = getattr(self, obj_fake_attr)

        self.rbac_policy.object_type = obj_type
        self.rbac_policy.object_id = obj_fake.id
        arglist = [
            '--type',
            obj_type,
            '--action',
            self.rbac_policy.action,
            '--target-project',
            self.rbac_policy.target_project_id,
            obj_fake.name,
        ]
        verifylist = [
            ('type', obj_type),
            ('action', self.rbac_policy.action),
            ('target_project', self.rbac_policy.target_project_id),
            ('rbac_object', obj_fake.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_rbac_policy.assert_called_with(
            **{
                'object_id': obj_fake.id,
                'object_type': obj_type,
                'action': self.rbac_policy.action,
                'target_tenant': self.rbac_policy.target_project_id,
            }
        )
        self.data = [
            self.rbac_policy.action,
            self.rbac_policy.id,
            obj_fake.id,
            obj_type,
            self.rbac_policy.project_id,
            self.rbac_policy.target_project_id,
        ]
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestDeleteNetworkRBAC(TestNetworkRBAC):
    rbac_policies = network_fakes.create_network_rbacs(count=2)

    def setUp(self):
        super().setUp()
        self.network_client.delete_rbac_policy.return_value = None
        self.network_client.find_rbac_policy = network_fakes.get_network_rbacs(
            rbac_policies=self.rbac_policies
        )

        # Get the command object to test
        self.cmd = network_rbac.DeleteNetworkRBAC(self.app, None)

    def test_network_rbac_delete(self):
        arglist = [
            self.rbac_policies[0].id,
        ]
        verifylist = [
            ('rbac_policy', [self.rbac_policies[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_rbac_policy.assert_called_once_with(
            self.rbac_policies[0].id, ignore_missing=False
        )
        self.network_client.delete_rbac_policy.assert_called_once_with(
            self.rbac_policies[0]
        )
        self.assertIsNone(result)

    def test_multi_network_rbacs_delete(self):
        arglist = []
        verifylist = []

        for r in self.rbac_policies:
            arglist.append(r.id)
        verifylist = [
            ('rbac_policy', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for r in self.rbac_policies:
            calls.append(call(r))
        self.network_client.delete_rbac_policy.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_network_policies_delete_with_exception(self):
        arglist = [
            self.rbac_policies[0].id,
            'unexist_rbac_policy',
        ]
        verifylist = [
            ('rbac_policy', [self.rbac_policies[0].id, 'unexist_rbac_policy']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self.rbac_policies[0], exceptions.CommandError]
        self.network_client.find_rbac_policy.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 RBAC policies failed to delete.', str(e))

        self.network_client.find_rbac_policy.assert_any_call(
            self.rbac_policies[0].id, ignore_missing=False
        )
        self.network_client.find_rbac_policy.assert_any_call(
            'unexist_rbac_policy', ignore_missing=False
        )
        self.network_client.delete_rbac_policy.assert_called_once_with(
            self.rbac_policies[0]
        )


class TestListNetworkRABC(TestNetworkRBAC):
    # The network rbac policies going to be listed up.
    rbac_policies = network_fakes.create_network_rbacs(count=3)

    columns = (
        'ID',
        'Object Type',
        'Object ID',
    )
    columns_long = (
        'ID',
        'Object Type',
        'Object ID',
        'Action',
    )
    data = []
    for r in rbac_policies:
        data.append(
            (
                r.id,
                r.object_type,
                r.object_id,
            )
        )
    data_long = []
    for r in rbac_policies:
        data_long.append(
            (
                r.id,
                r.object_type,
                r.object_id,
                r.action,
            )
        )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = network_rbac.ListNetworkRBAC(self.app, None)

        self.network_client.rbac_policies.return_value = self.rbac_policies

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

    def test_network_rbac_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_rbac_list_pagination(self):
        arglist = [
            '--marker',
            self.rbac_policies[0].id,
            '--limit',
            '1',
        ]
        verifylist = [
            ('marker', self.rbac_policies[0].id),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_once_with(
            **{
                'marker': self.rbac_policies[0].id,
                'limit': 1,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_rbac_list_type_opt(self):
        arglist = [
            '--type',
            self.rbac_policies[0].object_type,
        ]
        verifylist = [('type', self.rbac_policies[0].object_type)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_with(
            **{'object_type': self.rbac_policies[0].object_type}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_rbac_list_action_opt(self):
        arglist = [
            '--action',
            self.rbac_policies[0].action,
        ]
        verifylist = [('action', self.rbac_policies[0].action)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_with(
            **{'action': self.rbac_policies[0].action}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_rbac_list_with_long(self):
        arglist = [
            '--long',
        ]

        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_with()
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_network_rbac_list_target_project_opt(self):
        arglist = [
            '--target-project',
            self.rbac_policies[0].target_project_id,
        ]
        verifylist = [
            ('target_project', self.rbac_policies[0].target_project_id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.rbac_policies.assert_called_with(
            **{'target_project_id': self.project.id}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestSetNetworkRBAC(TestNetworkRBAC):
    project = sdk_fakes.generate_fake_resource(_project.Project)
    rbac_policy = network_fakes.create_one_network_rbac(
        attrs={'target_tenant': project.id}
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = network_rbac.SetNetworkRBAC(self.app, None)

        self.network_client.find_rbac_policy.return_value = self.rbac_policy

        self.network_client.update_rbac_policy.return_value = None
        self.identity_sdk_client.find_project.return_value = self.project

    def test_network_rbac_set_nothing(self):
        arglist = [
            self.rbac_policy.id,
        ]
        verifylist = [
            ('rbac_policy', self.rbac_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_rbac_policy.assert_called_once_with(
            self.rbac_policy.id, ignore_missing=False
        )
        attrs = {}
        self.network_client.update_rbac_policy.assert_called_once_with(
            self.rbac_policy, **attrs
        )
        self.assertIsNone(result)

    def test_network_rbac_set(self):
        arglist = [
            '--target-project',
            self.project.id,
            self.rbac_policy.id,
        ]
        verifylist = [
            ('target_project', self.project.id),
            ('rbac_policy', self.rbac_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_rbac_policy.assert_called_once_with(
            self.rbac_policy.id, ignore_missing=False
        )
        attrs = {'target_tenant': self.project.id}
        self.network_client.update_rbac_policy.assert_called_once_with(
            self.rbac_policy, **attrs
        )
        self.assertIsNone(result)


class TestShowNetworkRBAC(TestNetworkRBAC):
    rbac_policy = network_fakes.create_one_network_rbac()

    columns = (
        'action',
        'id',
        'object_id',
        'object_type',
        'project_id',
        'target_project_id',
    )

    data = [
        rbac_policy.action,
        rbac_policy.id,
        rbac_policy.object_id,
        rbac_policy.object_type,
        rbac_policy.project_id,
        rbac_policy.target_project_id,
    ]

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = network_rbac.ShowNetworkRBAC(self.app, None)

        self.network_client.find_rbac_policy.return_value = self.rbac_policy

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_rbac_show_all_options(self):
        arglist = [
            self.rbac_policy.id,
        ]
        verifylist = [
            ('rbac_policy', self.rbac_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_rbac_policy.assert_called_with(
            self.rbac_policy.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import qos_rule_type as _qos_rule_type
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _


def _get_columns(
    item: _qos_rule_type.QoSRuleType,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        "type": "rule_type_name",
        "drivers": "drivers",
    }
    hidden_columns = ["id", "location", "name", 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class ListNetworkQosRuleType(command.Lister):
    _description = _("List QoS rule types")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        supported = parser.add_mutually_exclusive_group()
        supported.add_argument(
            '--all-supported',
            action='store_true',
            help=_(
                "List all the QoS rule types supported by any loaded "
                "mechanism drivers (the union of all sets of supported "
                "rules)"
            ),
        )
        supported.add_argument(
            '--all-rules',
            action='store_true',
            help=_(
                "List all QoS rule types implemented in Neutron QoS driver"
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        columns = ('type',)
        column_headers = ('Type',)

        filters = {}
        if parsed_args.marker is not None:
            filters['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            filters['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            filters['max_items'] = parsed_args.max_items
        if parsed_args.all_supported:
            filters['all_supported'] = True
        elif parsed_args.all_rules:
            filters['all_rules'] = True
        data = client.qos_rule_types(**filters)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class ShowNetworkQosRuleType(command.ShowOne):
    _description = _("Show details about supported QoS rule type")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rule_type',
            metavar="<qos-rule-type-name>",
            help=_("Name of QoS rule type"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.get_qos_rule_type(parsed_args.rule_type)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return display_columns, data



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import ClassVar
import uuid

from openstackclient.tests.functional.volume.v3 import common


class VolumeGroupTests(common.BaseVolumeTests):
    """Functional tests for volume group."""

    API_VERSION = '3.13'
    GROUP_TYPE_NAME = uuid.uuid4().hex
    GROUP_TYPE_ID: ClassVar[str]
    VOLUME_TYPE_ID: ClassVar[str]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cmd_output = cls.openstack(
            '--os-volume-api-version '
            + cls.API_VERSION
            + ' volume group type create '
            + cls.GROUP_TYPE_NAME,
            parse_output=True,
        )
        cls.GROUP_TYPE_ID = cmd_output['ID']

        volume_type_name = uuid.uuid4().hex
        cmd_output = cls.openstack(
            'volume type create ' + volume_type_name,
            parse_output=True,
        )
        cls.VOLUME_TYPE_ID = cmd_output['id']

    @classmethod
    def tearDownClass(cls):
        try:
            raw_output = cls.openstack(
                'volume type delete ' + cls.VOLUME_TYPE_ID
            )
            cls.assertOutput('', raw_output)
            raw_output = cls.openstack(
                '--os-volume-api-version '
                + cls.API_VERSION
                + ' volume group type delete '
                + cls.GROUP_TYPE_NAME
            )
            cls.assertOutput('', raw_output)
        finally:
            super().tearDownClass()

    def test_volume_group(self):
        # create volume group
        name = uuid.uuid4().hex
        description = 'description-' + uuid.uuid4().hex
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group create '
            + '--volume-group-type '
            + self.GROUP_TYPE_NAME
            + ' --volume-type '
            + self.VOLUME_TYPE_ID
            + ' --name '
            + name
            + ' --description '
            + description,
            parse_output=True,
        )
        group_id = cmd_output['ID']
        self.addCleanup(
            self.wait_for_delete,
            '--os-volume-api-version ' + self.API_VERSION + ' volume group',
            group_id,
            name_field='ID',
        )
        self.addCleanup(
            self.openstack,
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group delete '
            + group_id,
            fail_ok=True,
        )
        self.assertIsNotNone(group_id)
        self.assertEqual(name, cmd_output['Name'])
        self.assertEqual(description, cmd_output['Description'])
        self.assertEqual(self.GROUP_TYPE_ID, cmd_output['Group Type'])
        self.assertIn(self.VOLUME_TYPE_ID, cmd_output['Volume Types'])

        # show volume group
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group show '
            + group_id,
            parse_output=True,
        )
        self.assertEqual(group_id, cmd_output['ID'])
        self.assertEqual(name, cmd_output['Name'])
        self.assertEqual(description, cmd_output['Description'])

        # list volume group
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group list',
            parse_output=True,
        )
        self.assertIn(group_id, [group['ID'] for group in cmd_output])
        self.assertIn(name, [group['Name'] for group in cmd_output])

        # set volume group
        new_name = uuid.uuid4().hex
        new_description = 'description-' + uuid.uuid4().hex
        self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group set '
            + '--name '
            + new_name
            + ' --description '
            + new_description
            + ' '
            + group_id,
        )

        # show updated volume group
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group show '
            + group_id,
            parse_output=True,
        )
        self.assertEqual(group_id, cmd_output['ID'])
        self.assertEqual(new_name, cmd_output['Name'])
        self.assertEqual(new_description, cmd_output['Description'])

        # delete volume group
        raw_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' volume group delete '
            + group_id,
        )
        self.assertOutput('', raw_output)
        self.wait_for_delete(
            '--os-volume-api-version ' + self.API_VERSION + ' volume group',
            group_id,
            name_field='ID',
        )



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
from unittest import mock
import uuid

from keystoneauth1 import access
from keystoneauth1 import fixture
from openstack.identity import v2 as identity_v2

from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils


project_id = '8-9-64'
project_name = 'beatles'
project_description = 'Fab Four'

PROJECT = {
    'id': project_id,
    'name': project_name,
    'description': project_description,
    'enabled': True,
}

PROJECT_2 = {
    'id': project_id + '-2222',
    'name': project_name + ' reprise',
    'description': project_description + 'plus four more',
    'enabled': True,
}

role_id = '1'
role_name = 'boss'

ROLE = {
    'id': role_id,
    'name': role_name,
}

ROLE_2 = {
    'id': '2',
    'name': 'bigboss',
}

service_id = '1925-10-11'
service_name = 'elmore'
service_description = 'Leonard, Elmore, rip'
service_type = 'author'

SERVICE = {
    'id': service_id,
    'name': service_name,
    'description': service_description,
    'type': service_type,
}

user_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
user_name = 'paul'
user_description = 'Sir Paul'
user_email = 'paul@applecorps.com'

USER = {
    'id': user_id,
    'name': user_name,
    'tenantId': project_id,
    'email': user_email,
    'enabled': True,
}

token_expires = '2016-09-05T18:04:52+0000'
token_id = 'token-id-' + uuid.uuid4().hex

TOKEN = {
    'expires': token_expires,
    'id': token_id,
    'tenant_id': 'project-id',
    'user_id': 'user-id',
}

UNSCOPED_TOKEN = {
    'expires': token_expires,
    'id': token_id,
    'user_id': 'user-id',
}

endpoint_name = service_name
endpoint_adminurl = 'https://admin.example.com/v2/UUID'
endpoint_region = 'RegionOne'
endpoint_internalurl = 'https://internal.example.com/v2/UUID'
endpoint_type = service_type
endpoint_id = '11b41ee1b00841128b7333d4bf1a6140'
endpoint_publicurl = 'https://public.example.com/v2/UUID'
endpoint_service_id = service_id

ENDPOINT = {
    'service_name': endpoint_name,
    'adminurl': endpoint_adminurl,
    'region': endpoint_region,
    'internalurl': endpoint_internalurl,
    'service_type': endpoint_type,
    'id': endpoint_id,
    'publicurl': endpoint_publicurl,
    'service_id': endpoint_service_id,
}


def fake_auth_ref(fake_token, fake_service=None):
    """Create an auth_ref using keystoneauth's fixtures"""
    token_copy = copy.deepcopy(fake_token)
    token_copy['token_id'] = token_copy.pop('id')
    token = fixture.V2Token(**token_copy)
    # An auth_ref is actually an access info object
    auth_ref = access.create(body=token)

    # Create a service catalog
    if fake_service:
        service = token.add_service(
            fake_service.type,
            fake_service.name,
        )
        # TODO(dtroyer): Add an 'id' element to KSA's _Service fixure
        service['id'] = fake_service.id
        for e in fake_service.endpoints:
            # KSA's _Service fixture copies publicURL to internalURL and
            # adminURL if they do not exist.  Soooo helpful...
            internal = e.get('internalURL', None)
            admin = e.get('adminURL', None)
            region = e.get('region_id') or e.get('region', '<none>')
            endpoint = service.add_endpoint(
                public=e['publicURL'],
                internal=internal,
                admin=admin,
                region=region,
            )
            # ...so undo that helpfulness
            if not internal:
                endpoint['internalURL'] = None
            if not admin:
                endpoint['adminURL'] = None

    return auth_ref


class FakeIdentityClient:
    def __init__(self, **kwargs):
        self.roles = mock.Mock()
        self.roles.resource_class = fakes.FakeResource(None, {})
        self.services = mock.Mock()
        self.services.resource_class = fakes.FakeResource(None, {})
        self.tenants = mock.Mock()
        self.tenants.resource_class = fakes.FakeResource(None, {})
        self.tokens = mock.Mock()
        self.tokens.resource_class = fakes.FakeResource(None, {})
        self.users = mock.Mock()
        self.users.resource_class = fakes.FakeResource(None, {})
        self.endpoints = mock.Mock()
        self.endpoints.resource_class = fakes.FakeResource(None, {})
        self.extensions = mock.Mock()
        self.extensions.resource_class = fakes.FakeResource(None, {})

    def __getattr__(self, name):
        # Map v3 'projects' back to v2 'tenants'
        if name == "projects":
            return self.tenants
        else:
            raise AttributeError(name)


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.identity = FakeIdentityClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.identity_client = self.app.client_manager.identity

        # TODO(stephenfin): Rename to 'identity_client' once all commands are
        # migrated to SDK
        self.app.client_manager.sdk_connection.identity = mock.Mock(
            identity_v2.Proxy
        )
        self.identity_sdk_client = (
            self.app.client_manager.sdk_connection.identity
        )


class TestIdentity(
    FakeClientMixin,
    utils.TestCommand,
): ...


class FakeCatalog:
    """Fake one or more catalog."""

    @staticmethod
    def create_catalog(attrs=None):
        """Create a fake catalog.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object with id, name, type and so on.
        """
        attrs = attrs or {}

        # Set default attributes.
        catalog_info = {
            'id': 'service-id-' + uuid.uuid4().hex,
            'type': 'compute',
            'name': 'supernova',
            'endpoints': [
                {
                    'region': 'one',
                    'publicURL': 'https://public.one.example.com',
                    'internalURL': 'https://internal.one.example.com',
                    'adminURL': 'https://admin.one.example.com',
                },
                {
                    'region': 'two',
                    'publicURL': 'https://public.two.example.com',
                    'internalURL': 'https://internal.two.example.com',
                    'adminURL': 'https://admin.two.example.com',
                },
                {
                    'region': None,
                    'publicURL': 'https://public.none.example.com',
                    'internalURL': 'https://internal.none.example.com',
                    'adminURL': 'https://admin.none.example.com',
                },
            ],
        }
        # Overwrite default attributes.
        catalog_info.update(attrs)

        catalog = fakes.FakeResource(
            info=copy.deepcopy(catalog_info), loaded=True
        )

        return catalog


class FakeProject:
    """Fake one or more project."""

    @staticmethod
    def create_one_project(attrs=None):
        """Create a fake project.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        project_info = {
            'id': 'project-id-' + uuid.uuid4().hex,
            'name': 'project-name-' + uuid.uuid4().hex,
            'description': 'project_description',
            'enabled': True,
        }
        project_info.update(attrs)

        project = fakes.FakeResource(
            info=copy.deepcopy(project_info), loaded=True
        )
        return project

    @staticmethod
    def create_projects(attrs=None, count=2):
        """Create multiple fake projects.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of projects to fake
        :return:
            A list of FakeResource objects faking the projects
        """
        projects = []
        for i in range(0, count):
            projects.append(FakeProject.create_one_project(attrs))

        return projects


class FakeEndpoint:
    """Fake one or more endpoint."""

    @staticmethod
    def create_one_endpoint(attrs=None):
        """Create a fake agent.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, region, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        endpoint_info = {
            'service_name': 'service-name-' + uuid.uuid4().hex,
            'adminurl': 'http://endpoint_adminurl',
            'region': 'endpoint_region',
            'internalurl': 'http://endpoint_internalurl',
            'service_type': 'service_type',
            'id': 'endpoint-id-' + uuid.uuid4().hex,
            'publicurl': 'http://endpoint_publicurl',
            'service_id': 'service-name-' + uuid.uuid4().hex,
        }
        endpoint_info.update(attrs)

        endpoint = fakes.FakeResource(
            info=copy.deepcopy(endpoint_info), loaded=True
        )
        return endpoint

    @staticmethod
    def create_endpoints(attrs=None, count=2):
        """Create multiple fake endpoints.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of endpoints to fake
        :return:
            A list of FakeResource objects faking the endpoints
        """
        endpoints = []
        for i in range(0, count):
            endpoints.append(FakeEndpoint.create_one_endpoint(attrs))

        return endpoints


class FakeService:
    """Fake one or more service."""

    @staticmethod
    def create_one_service(attrs=None):
        """Create a fake service.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, type, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        service_info = {
            'id': 'service-id-' + uuid.uuid4().hex,
            'name': 'service-name-' + uuid.uuid4().hex,
            'description': 'service_description',
            'type': 'service_type',
        }
        service_info.update(attrs)

        service = fakes.FakeResource(
            info=copy.deepcopy(service_info), loaded=True
        )
        return service

    @staticmethod
    def create_services(attrs=None, count=2):
        """Create multiple fake services.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of services to fake
        :return:
            A list of FakeResource objects faking the services
        """
        services = []
        for i in range(0, count):
            services.append(FakeService.create_one_service(attrs))

        return services


class FakeRole:
    """Fake one or more role."""

    @staticmethod
    def create_one_role(attrs=None):
        """Create a fake role.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        role_info = {
            'id': 'role-id' + uuid.uuid4().hex,
            'name': 'role-name' + uuid.uuid4().hex,
        }
        role_info.update(attrs)

        role = fakes.FakeResource(info=copy.deepcopy(role_info), loaded=True)
        return role

    @staticmethod
    def create_roles(attrs=None, count=2):
        """Create multiple fake roles.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of roles to fake
        :return:
            A list of FakeResource objects faking the roles
        """
        roles = []
        for i in range(0, count):
            roles.append(FakeRole.create_one_role(attrs))

        return roles


class FakeUser:
    """Fake one or more user."""

    @staticmethod
    def create_one_user(attrs=None):
        """Create a fake user.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """
        attrs = attrs or {}

        # set default attributes.
        user_info = {
            'id': 'user-id-' + uuid.uuid4().hex,
            'name': 'user-name-' + uuid.uuid4().hex,
            'tenantId': 'project-id-' + uuid.uuid4().hex,
            'email': 'admin@openstack.org',
            'enabled': True,
        }
        user_info.update(attrs)

        user = fakes.FakeResource(info=copy.deepcopy(user_info), loaded=True)
        return user

    @staticmethod
    def create_users(attrs=None, count=2):
        """Create multiple fake users.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of users to fake
        :return:
            A list of FakeResource objects faking the users
        """
        users = []
        for i in range(0, count):
            users.append(FakeUser.create_one_user(attrs))

        return users



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
from typing import ClassVar
import uuid

from openstackclient.tests.functional.network.v2 import common


class FloatingIpTests(common.NetworkTests):
    """Functional tests for floating ip"""

    EXTERNAL_NETWORK_NAME: ClassVar[str]
    EXTERNAL_NETWORK_ID: ClassVar[str]
    PRIVATE_NETWORK_NAME: ClassVar[str]
    PRIVATE_NETWORK_ID: ClassVar[str]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if cls.haz_network:
            # Create common networks that all tests share
            cls.EXTERNAL_NETWORK_NAME = uuid.uuid4().hex
            cls.PRIVATE_NETWORK_NAME = uuid.uuid4().hex

            # Create a network for the floating ip
            json_output = cls.openstack(
                'network create ' + '--external ' + cls.EXTERNAL_NETWORK_NAME,
                parse_output=True,
            )
            cls.EXTERNAL_NETWORK_ID = json_output["id"]

            # Create a private network for the port
            json_output = cls.openstack(
                'network create ' + cls.PRIVATE_NETWORK_NAME,
                parse_output=True,
            )
            cls.PRIVATE_NETWORK_ID = json_output["id"]

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_network:
                del_output = cls.openstack(
                    'network delete '
                    + cls.EXTERNAL_NETWORK_NAME
                    + ' '
                    + cls.PRIVATE_NETWORK_NAME
                )
                cls.assertOutput('', del_output)
        finally:
            super().tearDownClass()

    def setUp(self):
        super().setUp()

        # Verify setup
        self.assertIsNotNone(self.EXTERNAL_NETWORK_ID)
        self.assertIsNotNone(self.PRIVATE_NETWORK_ID)

    def _create_subnet(self, network_name, subnet_name):
        subnet_id = None

        # Try random subnet range for subnet creating
        # Because we can not determine ahead of time what subnets are
        # already in use, possibly by another test running in parallel,
        # try 4 times
        for i in range(4):
            # Make a random subnet
            subnet = (
                ".".join(map(str, (random.randint(0, 223) for _ in range(3))))
                + ".0/26"
            )
            try:
                # Create a subnet for the network
                json_output = self.openstack(
                    'subnet create '
                    + '--network '
                    + network_name
                    + ' '
                    + '--subnet-range '
                    + subnet
                    + ' '
                    + subnet_name,
                    parse_output=True,
                )
                self.assertIsNotNone(json_output["id"])
                subnet_id = json_output["id"]
            except Exception:
                if i == 3:
                    # raise the exception at the last time
                    raise
                pass
            else:
                # break and no longer retry if create successfully
                break
        return subnet_id

    def test_floating_ip_delete(self):
        """Test create, delete multiple"""

        # Subnets must exist even if not directly referenced here
        ext_subnet_id = self._create_subnet(
            self.EXTERNAL_NETWORK_NAME, "ext-test-delete"
        )
        self.addCleanup(self.openstack, 'subnet delete ' + ext_subnet_id)

        json_output = self.openstack(
            'floating ip create '
            + '--description aaaa '
            + self.EXTERNAL_NETWORK_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        ip1 = json_output["id"]
        self.assertEqual(
            'aaaa',
            json_output["description"],
        )

        json_output = self.openstack(
            'floating ip create '
            + '--description bbbb '
            + self.EXTERNAL_NETWORK_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        ip2 = json_output["id"]
        self.assertEqual(
            'bbbb',
            json_output["description"],
        )

        # Clean up after ourselves
        del_output = self.openstack('floating ip delete ' + ip1 + ' ' + ip2)
        self.assertOutput('', del_output)

        self.assertIsNotNone(json_output["floating_network_id"])

    def test_floating_ip_list(self):
        """Test create defaults, list filters, delete"""

        # Subnets must exist even if not directly referenced here
        ext_subnet_id = self._create_subnet(
            self.EXTERNAL_NETWORK_NAME, "ext-test-delete"
        )
        self.addCleanup(self.openstack, 'subnet delete ' + ext_subnet_id)

        json_output = self.openstack(
            'floating ip create '
            + '--description aaaa '
            + self.EXTERNAL_NETWORK_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        ip1 = json_output["id"]
        self.addCleanup(self.openstack, 'floating ip delete ' + ip1)
        self.assertEqual(
            'aaaa',
            json_output["description"],
        )
        self.assertIsNotNone(json_output["floating_network_id"])
        fip1 = json_output["floating_ip_address"]

        json_output = self.openstack(
            'floating ip create '
            + '--description bbbb '
            + self.EXTERNAL_NETWORK_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        ip2 = json_output["id"]
        self.addCleanup(self.openstack, 'floating ip delete ' + ip2)
        self.assertEqual(
            'bbbb',
            json_output["description"],
        )
        self.assertIsNotNone(json_output["floating_network_id"])
        fip2 = json_output["floating_ip_address"]

        # Test list
        json_output = self.openstack(
            'floating ip list',
            parse_output=True,
        )
        fip_map = {
            item.get('ID'): item.get('Floating IP Address')
            for item in json_output
        }
        # self.assertEqual(item_map, json_output)
        self.assertIn(ip1, fip_map.keys())
        self.assertIn(ip2, fip_map.keys())
        self.assertIn(fip1, fip_map.values())
        self.assertIn(fip2, fip_map.values())

        # Test list --long
        json_output = self.openstack(
            'floating ip list ' + '--long',
            parse_output=True,
        )
        fip_map = {
            item.get('ID'): item.get('Floating IP Address')
            for item in json_output
        }
        self.assertIn(ip1, fip_map.keys())
        self.assertIn(ip2, fip_map.keys())
        self.assertIn(fip1, fip_map.values())
        self.assertIn(fip2, fip_map.values())
        desc_map = {
            item.get('ID'): item.get('Description') for item in json_output
        }
        self.assertIn('aaaa', desc_map.values())
        self.assertIn('bbbb', desc_map.values())

        # TODO(dtroyer): add more filter tests

        json_output = self.openstack(
            'floating ip show ' + ip1,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        self.assertEqual(
            ip1,
            json_output["id"],
        )
        self.assertEqual(
            'aaaa',
            json_output["description"],
        )
        self.assertIsNotNone(json_output["floating_network_id"])
        self.assertEqual(
            fip1,
            json_output["floating_ip_address"],
        )

    def test_floating_ip_set_and_unset_port(self):
        """Test Floating IP Set and Unset port"""

        # Subnets must exist even if not directly referenced here
        ext_subnet_id = self._create_subnet(
            self.EXTERNAL_NETWORK_NAME, "ext-test-delete"
        )
        self.addCleanup(self.openstack, 'subnet delete ' + ext_subnet_id)
        priv_subnet_id = self._create_subnet(
            self.PRIVATE_NETWORK_NAME, "priv-test-delete"
        )
        self.addCleanup(self.openstack, 'subnet delete ' + priv_subnet_id)

        self.ROUTER = uuid.uuid4().hex
        self.PORT_NAME = uuid.uuid4().hex

        json_output = self.openstack(
            'floating ip create '
            + '--description aaaa '
            + self.EXTERNAL_NETWORK_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        ip1 = json_output["id"]
        self.addCleanup(self.openstack, 'floating ip delete ' + ip1)
        self.assertEqual(
            'aaaa',
            json_output["description"],
        )

        json_output = self.openstack(
            'port create '
            + '--network '
            + self.PRIVATE_NETWORK_NAME
            + ' '
            + '--fixed-ip subnet='
            + priv_subnet_id
            + ' '
            + self.PORT_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        port_id = json_output["id"]

        json_output = self.openstack(
            'router create ' + self.ROUTER,
            parse_output=True,
        )
        self.assertIsNotNone(json_output["id"])
        self.addCleanup(self.openstack, 'router delete ' + self.ROUTER)

        self.openstack('router add port ' + self.ROUTER + ' ' + port_id)

        self.openstack(
            'router set '
            + '--external-gateway '
            + self.EXTERNAL_NETWORK_NAME
            + ' '
            + self.ROUTER
        )
        self.addCleanup(
            self.openstack,
            'router unset --external-gateway ' + self.ROUTER,
        )
        self.addCleanup(
            self.openstack,
            'router remove port ' + self.ROUTER + ' ' + port_id,
        )

        self.openstack('floating ip set ' + '--port ' + port_id + ' ' + ip1)
        self.addCleanup(
            self.openstack,
            'floating ip unset --port ' + ip1,
        )

        json_output = self.openstack(
            'floating ip show ' + ip1,
            parse_output=True,
        )

        self.assertEqual(
            port_id,
            json_output["port_id"],
        )
